@@ -3,64 +3,64 @@ The code submodule for Jaypore CI.
 """
 import time
 import os
-import re
-from enum import Enum
+import subprocess
 from itertools import product
-from collections import defaultdict, namedtuple
+from collections import defaultdict
 from typing import List, Union, Callable
 from contextlib import contextmanager
 
 import structlog
 import pendulum
 
-from jaypore_ci import gitea, docker
-from jaypore_ci.interfaces import Remote, Executor, TriggerFailed
-from jaypore_ci.logging import logger, jaypore_logs
+from jaypore_ci import remotes, executors, reporters, repos
+from jaypore_ci.interfaces import (
+    Remote,
+    Executor,
+    Reporter,
+    TriggerFailed,
+    Status,
+    Repo,
+)
+from jaypore_ci.logging import logger
 
 TZ = "UTC"
 
 __all__ = ["Pipeline", "Job"]
 
 
-class Status(Enum):
-    "Each pipeline can be in any one of these statuses"
-    PENDING = 10
-    RUNNING = 30
-    FAILED = 40
-    PASSED = 50
-    TIMEOUT = 60
-    SKIPPED = 70
-
-
 # All of these statuses are considered "finished" statuses
 FIN_STATUSES = (Status.FAILED, Status.PASSED, Status.TIMEOUT, Status.SKIPPED)
-ansi_escape = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")
-
-
-def __node_mod__(nodes):
-    mod = 1
-    if len(nodes) > 5:
-        mod = 2
-    if len(nodes) > 10:
-        mod = 3
-    return mod
-
-
-def __clean_logs__(logs):
-    """
-    Clean logs so that they don't have HTML/ANSI color codes in them.
-    """
-    for old, new in [("<", r"\<"), (">", r"\>"), ("`", '"'), ("\r", "\n")]:
-        logs = logs.replace(old, new)
-    return [line.strip() for line in ansi_escape.sub("", logs).split("\n")]
+PREFIX = "JAYPORE_"
 
 
 class Job:  # pylint: disable=too-many-instance-attributes
     """
-    This is the fundamental building block.
-    Each job goes through a lifecycle defined by `Status` class.
+    This is the fundamental building block for running jobs.
+    Each job goes through a lifecycle defined by
+    :class:`~jaypore_ci.interfaces.Status`.
 
-    A job is run by an Executor as part of a Pipeline.
+    A job is run by an :class:`~jaypore_ci.interfaces.Executor` as part of a
+    :class:`~jaypore_ci.jci.Pipeline`.
+
+    It is never created manually. The correct way to create a job is to use
+    :meth:`~jaypore_ci.jci.Pipeline.job`.
+
+    :param name: The name for the job. Names must be unique across jobs and stages.
+    :param command: The command that we need to run for the job. It can be set
+        to `None` when `is_service` is True.
+    :param is_service: Is this job a service or not? Service jobs are assumed
+        to be :class:`~jaypore_ci.interfaces.Status.PASSED` as long as they start.
+        They are shut down when the entire pipeline has finished executing.
+    :param pipeline: The pipeline this job is associated with.
+    :param status: The :class:`~jaypore_ci.interfaces.Status` of this job.
+    :param image: What docker image to use for this job.
+    :param timeout: Defines how long a job is allowed to run before being
+        killed and marked as class:`~jaypore_ci.interfaces.Status.FAILED`.
+    :param env: A dictionary of environment variables to pass to the docker run command.
+    :param children: Defines which jobs depend on this job's output status.
+    :param parents: Defines which jobs need to pass before this job can be run.
+    :param stage: What stage the job belongs to. This stage name must exist so
+        that we can assign jobs to it.
     """
 
     def __init__(
@@ -82,6 +82,7 @@ class Job:  # pylint: disable=too-many-instance-attributes
         self.command = command
         self.image = image
         self.status = status
+        self.run_state = None
         self.timeout = timeout
         self.pipeline = pipeline
         self.env = env
@@ -107,7 +108,7 @@ class Job:  # pylint: disable=too-many-instance-attributes
             run_id=self.run_id,
         )
 
-    def update_report(self):
+    def update_report(self) -> str:
         """
         Update the status report. Usually called when a job changes some of
         it's internal state like when logs are updated or when status has
@@ -122,7 +123,14 @@ class Job:  # pylint: disable=too-many-instance-attributes
             Status.TIMEOUT: "warning",
             Status.SKIPPED: "warning",
         }[self.pipeline.get_status()]
-        self.pipeline.remote.publish(self.pipeline.render_report(), status)
+        report = self.pipeline.reporter.render(self.pipeline)
+        with open("/jaypore_ci/run/jaypore_ci.status.txt", "w", encoding="utf-8") as fl:
+            fl.write(report)
+        try:
+            self.pipeline.remote.publish(report, status)
+        except Exception as e:  # pylint: disable=broad-except
+            self.logging().exception(e)
+        return report
 
     def trigger(self):
         """
@@ -148,7 +156,7 @@ class Job:  # pylint: disable=too-many-instance-attributes
                         job_name=self.name,
                     )
                     logs = job_run.stdout.decode()
-                    self.logs["stdout"] = __clean_logs__(logs)
+                    self.logs["stdout"] = reporters.clean_logs(logs)
                     self.status = Status.FAILED
         else:
             self.logging().info("Trigger called but job already running")
@@ -161,20 +169,24 @@ class Job:  # pylint: disable=too-many-instance-attributes
         """
         if isinstance(self.command, str) and self.run_id is not None:
             self.logging().debug("Checking job run")
-            is_running, exit_code, logs = self.pipeline.executor.get_status(self.run_id)
+            self.run_state = self.pipeline.executor.get_status(self.run_id)
             self.last_check = pendulum.now(TZ)
             self.logging().debug(
-                "Job run status found", is_running=is_running, exit_code=exit_code
+                "Job run status found",
+                is_running=self.run_state.is_running,
+                exit_code=self.run_state.exit_code,
             )
-            if is_running:
+            if self.run_state.is_running:
                 self.status = Status.RUNNING if not self.is_service else Status.PASSED
             else:
-                self.status = Status.PASSED if exit_code == 0 else Status.FAILED
-            self.logs["stdout"] = __clean_logs__(logs)
+                self.status = (
+                    Status.PASSED if self.run_state.exit_code == 0 else Status.FAILED
+                )
+            self.logs["stdout"] = reporters.clean_logs(self.run_state.logs)
             if with_update_report:
                 self.update_report()
 
-    def is_complete(self):
+    def is_complete(self) -> bool:
         """
         Is this job complete? It could have passed/ failed etc.
         We no longer need to check for updates in a complete job.
@@ -183,40 +195,71 @@ class Job:  # pylint: disable=too-many-instance-attributes
 
     def get_env(self):
         """
-        Gets the environment variables for a given job by interpolating it with
-        the pipeline's environment.
+        Gets the environment variables for a given job.
+        Order of precedence for setting values is:
+
+        1. Pipeline
+        2. Stage
+        3. Job
         """
-        return {**os.environ, **self.pipeline.pipe_kwargs.get("env", {}), **self.env}
+        return {
+            **{
+                k[len(PREFIX) :]: v
+                for k, v in os.environ.items()
+                if k.startswith(PREFIX)
+            },
+            **self.pipeline.pipe_kwargs.get("env", {}),
+            **self.env,
+        }
 
 
 class Pipeline:  # pylint: disable=too-many-instance-attributes
     """
     A pipeline acts as a controlling/organizing mechanism for multiple jobs.
 
-    - Each pipeline has stages. A default stage of 'Pipeline' is always available.
-    - Stages are executed in order. Execution proceeds to the next stage ONLY
-      if all jobs in a stage have passed.
-    - Jobs can be defined inside stages.
+    :param repo         : Provides information about the codebase.
+    :param reporter     : Provides reports based on the state of the pipeline.
+    :param remote       : Allows us to publish reports to somewhere like gitea/email.
+    :param executor     : Runs the specified jobs.
+    :param poll_interval: Defines how frequently (in seconds) to check the
+        pipeline status and publish a report.
     """
 
     def __init__(  # pylint: disable=too-many-arguments
         self,
+        *,
+        repo: Repo = None,
         remote: Remote = None,
         executor: Executor = None,
-        *,
-        graph_direction: str = "TB",
+        reporter: Reporter = None,
+        poll_interval: int = 10,
         **kwargs,
     ) -> "Pipeline":
         self.jobs = {}
         self.services = []
         self.should_pass_called = set()
-        self.remote = remote if remote is not None else gitea.Gitea.from_env()
-        self.executor = executor if executor is not None else docker.Docker()
-        self.graph_direction = graph_direction
-        self.executor.set_pipeline(self)
+        self.repo = repo if repo is not None else repos.Git.from_env()
+        self.remote = (
+            remote
+            if remote is not None
+            else remotes.gitea.Gitea.from_env(repo=self.repo)
+        )
+        self.executor = executor if executor is not None else executors.docker.Docker()
+        self.reporter = reporter if reporter is not None else reporters.text.Text()
+        self.poll_interval = poll_interval
         self.stages = ["Pipeline"]
+        self.pipe_id = (
+            subprocess.check_output(
+                "cat /proc/self/cgroup | grep name= | awk -F/ '{print $3}'",
+                shell=True,
+                stderr=subprocess.STDOUT,
+            )
+            .decode()
+            .strip()
+        )
+        self.executor.set_pipeline(self)
         # ---
-        kwargs["image"] = kwargs.get("image", "arjoonn/jaypore_ci:latest")
+        kwargs["image"] = kwargs.get("image", "arjoonn/jci:latest")
         kwargs["timeout"] = kwargs.get("timeout", 15 * 60)
         kwargs["env"] = kwargs.get("env", {})
         kwargs["stage"] = "Pipeline"
@@ -247,7 +290,7 @@ class Pipeline:  # pylint: disable=too-many-instance-attributes
         self.remote.__exit__(exc_type, exc_value, traceback)
         return False
 
-    def get_status(self):
+    def get_status(self) -> Status:
         """
         Calculates a pipeline's status based on the status of it's jobs.
         """
@@ -258,17 +301,18 @@ class Pipeline:  # pylint: disable=too-many-instance-attributes
         for service in self.services:
             service.check_job(with_update_report=False)
         if service is not None:
-            service.check_job()
-        has_pending = True
+            service.check_job(with_update_report=False)
+        has_pending = False
         for job in self.jobs.values():
             job.check_job(with_update_report=False)
-            if job.is_complete():
-                has_pending = False
+            if not job.is_complete():
+                has_pending = True
+            else:
                 if job.status != Status.PASSED:
                     return Status.FAILED
         return Status.PENDING if has_pending else Status.PASSED
 
-    def get_status_dot(self):
+    def get_status_dot(self) -> str:
         """
         Get's the status dot for the pipeline.
         """
@@ -280,113 +324,6 @@ class Pipeline:  # pylint: disable=too-many-instance-attributes
             return "🔵"
         return "🟡"
 
-    def render_report(self):
-        """
-        Returns a markdown report for a given pipeline.
-
-        It will include a mermaid graph and a collapsible list of logs for each
-        job.
-        """
-        return f"""
-<details>
-    <summary>JayporeCi: {self.get_status_dot()} {self.remote.sha[:10]}</summary>
-
-{self.__render_graph__()}
-{self.__render_logs__()}
-
-</details>"""
-
-    def __render_graph__(self) -> str:  # pylint: disable=too-many-locals
-        """
-        Render a mermaid graph given the jobs in the pipeline.
-        """
-        st_map = {
-            Status.PENDING: "pending",
-            Status.RUNNING: "running",
-            Status.FAILED: "failed",
-            Status.PASSED: "passed",
-            Status.TIMEOUT: "timeout",
-            Status.SKIPPED: "skipped",
-        }
-        mermaid = f"""
-```mermaid
-flowchart {self.graph_direction}
-"""
-        for stage in self.stages:
-            nodes, edges = set(), set()
-            for job in self.jobs.values():
-                if job.stage != stage:
-                    continue
-                nodes.add(job.name)
-                edges |= {(p, job.name) for p in job.parents}
-            mermaid += f"""
-            subgraph {stage}
-                direction {self.graph_direction}
-            """
-            ref = {n: f"{stage}_{i}" for i, n in enumerate(nodes)}
-            # If there are too many nodes, scatter them with different length arrows
-            mod = __node_mod__([n for n in nodes if not self.jobs[n].parents])
-            for i, n in enumerate(nodes):
-                n = self.jobs[n]
-                if n.parents:
-                    continue
-                arrow = "." * ((i % mod) + 1)
-                arrow = f"-{arrow}->"
-                mermaid += f"""
-                s_{stage}(( )) {arrow} {ref[n.name]}({n.name}):::{st_map[n.status]}"""
-            mod = __node_mod__([n for n in nodes if self.jobs[n].parents])
-            for i, (a, b) in enumerate(edges):
-                a, b = self.jobs[a], self.jobs[b]
-                arrow = "." * ((i % mod) + 1)
-                arrow = f"-{arrow}->"
-                mermaid += f"""
-                {ref[a.name]}({a.name}):::{st_map[a.status]} {arrow} {ref[b.name]}({b.name}):::{st_map[b.status]}"""
-            mermaid += """
-            end
-            """
-        for s1, s2 in zip(self.stages, self.stages[1:]):
-            mermaid += f"""
-            {s1} ---> {s2}
-            """
-        mermaid += """
-
-            classDef pending fill:#aaa, color:black, stroke:black,stroke-width:2px,stroke-dasharray: 5 5;
-            classDef skipped fill:#aaa, color:black, stroke:black,stroke-width:2px;
-            classDef assigned fill:#ddd, color:black, stroke:black,stroke-width:2px;
-            classDef running fill:#bae1ff,color:black,stroke:black,stroke-width:2px,stroke-dasharray: 5 5;
-            classDef passed fill:#88d8b0, color:black, stroke:black;
-            classDef failed fill:#ff6f69, color:black, stroke:black;
-            classDef timeout fill:#ffda9e, color:black, stroke:black;
-``` """
-        return mermaid
-
-    def __render_logs__(self):
-        """
-        Collect all pipeline logs and render into a single collapsible text.
-        """
-        all_logs = []
-        fake_job = namedtuple("fake_job", "name logs")(
-            "JayporeCi",
-            {"stdout": __clean_logs__("\n".join(jaypore_logs))},
-        )
-        for job in [fake_job] + list(self.jobs.values()):
-            job_log = []
-            for logname, stream in job.logs.items():
-                job_log += [f"============== {logname} ============="]
-                job_log += [line.strip() for line in stream]
-            if job_log:
-                all_logs += [
-                    "- <details>",
-                    f"    <summary>Logs: {job.name}</summary>",
-                    "",
-                    "    ```",
-                    *[f"    {line}" for line in job_log],
-                    "    ```",
-                    "",
-                    "  </details>",
-                ]
-        return "\n".join(all_logs)
-
     def job(
         self,
         name: str,
@@ -396,21 +333,20 @@ flowchart {self.graph_direction}
         **kwargs,
     ) -> Job:
         """
-        Declare a job in this pipeline.
-
-        Jobs inherit their keyword arguments from the stage they are defined in
-        and the pipeline they are defined in.
-
-        Initially jobs are in a `PENDING` state.
+        Creates a :class:`~jaypore_ci.jci.Job` instance based on the
+        pipeline/stage that it is being defined in. See
+        :class:`~jaypore_ci.jci.Job` for details on what parameters can be
+        passed to the job.
         """
         depends_on = [] if depends_on is None else depends_on
         depends_on = [depends_on] if isinstance(depends_on, str) else depends_on
-        assert name not in self.jobs
+        assert name not in self.jobs, f"{name} already defined"
+        assert name not in self.stages, "Stage name cannot match a job's name"
         kwargs, job_kwargs = dict(self.pipe_kwargs), kwargs
         kwargs.update(self.stage_kwargs if self.stage_kwargs is not None else {})
         kwargs.update(job_kwargs)
         if not kwargs.get("is_service"):
-            assert command
+            assert command, f"Command: {command}"
         job = Job(
             name=name if name is not None else " ",
             command=command,
@@ -431,7 +367,8 @@ flowchart {self.graph_direction}
             self.services.append(job)
         return job
 
-    def env_matrix(self, **kwargs):
+    @classmethod
+    def env_matrix(cls, **kwargs):
         """
         Return a cartesian product of all the provided kwargs.
         """
@@ -439,21 +376,25 @@ flowchart {self.graph_direction}
         for values in product(*[kwargs[key] for key in keys]):
             yield dict(list(zip(keys, values)))
 
-    def run(self):
-        """
-        Run the pipeline. This is almost always called automatically when the
-        context of the pipeline declaration finishes.
-        """
-        # Ensure duplex connection between all nodes
+    def __ensure_duplex__(self):
         for name, job in self.jobs.items():
             for parent_name in job.parents:
                 parent = self.jobs[parent_name]
                 parent.children = list(sorted(set(parent.children).union(set([name]))))
+
+    def run(self):
+        """
+        Run the pipeline. This is always called automatically when the context
+        of the pipeline declaration finishes and so unless you are doing
+        something fancy you don't need to call this manually.
+        """
+        self.__ensure_duplex__()
         # Run stages one by one
+        job = None
         for stage in self.stages:
             # --- Trigger starting jobs
             jobs = {name: job for name, job in self.jobs.items() if job.stage == stage}
-            for name in {job.name for job in jobs.values() if job.parents}:
+            for name in {job.name for job in jobs.values() if not job.parents}:
                 jobs[name].trigger()
             # --- monitor and ensure all jobs run
             while not all(job.is_complete() for job in jobs.values()):
@@ -474,7 +415,7 @@ flowchart {self.graph_direction}
                         ):
                             job.status = Status.SKIPPED
                 job.check_job()
-                time.sleep(1)
+                time.sleep(self.poll_interval)
             # --- has this stage passed?
             if not all(
                 job.is_complete() and job.status == Status.PASSED
@@ -484,7 +425,9 @@ flowchart {self.graph_direction}
                 job.update_report()
                 break
         self.logging().error("Pipeline passed")
-        job.update_report()
+        if job is not None:
+            report = job.update_report()
+            self.logging().info("Report:", report=report)
 
     @contextmanager
     def stage(self, name, **kwargs):
